#include <stdlib.h>
#include <android/bitmap.h>

#include "ComicFinder.hpp"

JNIEXPORT bool JNICALL Java_com_example_comicfinder_MainActivity_processImage
    (JNIEnv *env, jobject me, jbyteArray src, jobject dst, jint width, jint height)
{
    int ret;
    unsigned char *p_dst;
    if ((ret = AndroidBitmap_lockPixels(env, dst, reinterpret_cast<void **>(&p_dst))) < 0) {
        return false;
    }

    unsigned char *p_src = reinterpret_cast<unsigned char *>(env->GetPrimitiveArrayCritical(src, NULL));
    if (NULL == p_src) {
        AndroidBitmap_unlockPixels(env, dst);
        return false;
    }

    const int threshold_black_and_gray = 60;
    const int threshold_gray_and_white = 120;
    const int threshold_edge = 32;

    int lum, fx, fy;
    unsigned char line_buf[(width+2)*3];
    unsigned char *line0 = line_buf;
    unsigned char *line1 = line0 + width+2;
    unsigned char *line2 = line1 + width+2;
    memcpy(line1+1, p_src, width);
    line1[0] = line1[1];
    line1[width+1] = line1[width];
    memcpy(line2+1, p_src+width, width);
    line2[0] = line2[1];
    line2[width+1] = line2[width];
    memcpy(line0, line1, width+2);

    for (int i = 0; i < height; i++) {
        for (int j = 0; j < width; j++) {
            lum = p_src[i*width+j];

            if (lum < threshold_black_and_gray) {
                lum = 0x00;
            }
            else if (lum < threshold_gray_and_white) {
                lum = ((i+j) % 2) == 0 ? 0x00 : 0xff;
            }
            else {
                lum = 0xff;
            }

            if (lum != 0x00) {
                fx = 0;
                fx += line0[j+0] * -1;
                fx += line1[j+0] * -2;
                fx += line2[j+0] * -1;
                fx += line0[j+2] *  1;
                fx += line1[j+2] *  2;
                fx += line2[j+2] *  1;
                fy = 0;
                fy += line0[j+0] * -1;
                fy += line0[j+1] * -2;
                fy += line0[j+2] * -1;
                fy += line2[j+0] *  1;
                fy += line2[j+1] *  2;
                fy += line2[j+2] *  1;
                lum = abs(fx) + abs(fy);
                lum = lum < threshold_edge ? 0xff : 0x00;
            }

            p_dst[(i*width+j)*4+0] = p_dst[(i*width+j)*4+1] = p_dst[(i*width+j)*4+2] = lum;
            p_dst[(i*width+j)*4+3] = 0xff;
        }

        // update line buffers.
        unsigned char *tmp = line0;
        line0 = line1;
        line1 = line2;
        line2 = tmp;
        memcpy(line2+1, p_src+(i+1 > height-1 ? height-1 : i+1)*width, width);
        line2[0] = line2[1];
        line2[width+1] = line2[width];
    }

    env->ReleasePrimitiveArrayCritical(src, p_src, 0);
    AndroidBitmap_unlockPixels(env, dst);

    return true;
}
